from itertools import zip_longest


def assert_bytes_equal(a, b):
    # type: (Union[bytes, bytearray], Union[bytes, bytearray]) -> None
    """
    Helper method to compare bytes with more helpful output.
    """
    __tracebackhide__ = True

    def is_bytes(x):
        # type: (Union[bytes, bytearray]) -> bool
        return isinstance(x, (bytes, bytearray))

    if not is_bytes(a) or not is_bytes(b):
        raise ValueError("assertBytesEqual requires two bytes objects!")

    if a != b:
        comparisons = []
        type_a = type(a)
        type_b = type(b)
        a = bytearray(a)
        b = bytearray(b)

        def char_repr(c):
            # type: (bytes) -> str
            if 0x1F < char_a < 0x80:
                # bytearray to prevent accidental pre-mature str conv
                # str to prevent b'' suffix in repr's output
                return repr(str(bytearray([char_a]).decode("ascii")))
            return "."

        for offset, (char_a, char_b) in enumerate(zip_longest(a, b)):
            comp, marker = ("==", "") if char_a == char_b else ("!=", ">>")

            # Using "zip_longest", overflows are marked as "None", which is
            # unambiguous in this case, but we need to handle these
            # separately from the main format string.
            if char_a is None:
                char_ab = char_ad = char_ah = char_ar = "?"
            else:
                char_ab = f"0b{char_a:08b}"
                char_ad = f"{char_a:3d}"
                char_ah = f"0x{char_a:02x}"
                char_ar = char_repr(char_a)
            if char_b is None:
                char_bb = char_bd = char_bh = char_br = "?"
            else:
                char_bb = f"0b{char_b:08b}"
                char_bd = f"{char_b:3d}"
                char_bh = f"0x{char_b:02x}"
                char_br = char_repr(char_b)
            comparisons.append(
                "{8:<3} Offset {0:4d}: "
                "{1:^10} {4} {5:^10} | "
                "{2:>3} {4} {6:>3} | "
                "{3:^4} {4} {7:^4} | {9:>3} {4} {10:>3}".format(
                    offset,
                    char_ab,
                    char_ad,
                    char_ah,
                    comp,
                    char_bb,
                    char_bd,
                    char_bh,
                    marker,
                    char_ar,
                    char_br,
                )
            )
        raise AssertionError(
            "Bytes differ!\n"
            + "type(a)=%s, type(b)=%s\n" % (type_a, type_b)
            + "\nIndividual bytes:\n"
            + "\n".join(comparisons)
        )



"""
We should be able to decode values without creating excessive copies in memory
"""

import pytest

import x690.types as t
from x690 import decode


@pytest.mark.parametrize(
    "data, index, expected, expected_nxt",
    [
        (b"padding\x02\x01\x01padding", 7, t.Integer(1), 10),
        (b"padding\x04\x03foopadding", 7, t.OctetString(b"foo"), 12),
        (
            b"pa\x00dding\x04\x80foo\x00\x00padding",
            8,
            t.OctetString(b"foo"),
            15,
        ),
    ],
)
def test_offset_decoding(data, index, expected, expected_nxt):
    """
    To work with non-copied data, we need to be able to decode at any point
    in the stream
    """
    value, next_tlv = decode(data, index)
    assert value == expected
    assert next_tlv == expected_nxt



# pylint: skip-file
import pytest

from x690 import types as t
from x690.util import TypeClass, TypeInfo, TypeNature

UNIVERSAL = TypeClass.UNIVERSAL
APPLICATION = TypeClass.APPLICATION
PRIVATE = TypeClass.PRIVATE
CONTEXT = TypeClass.CONTEXT
PRIMITIVE = TypeNature.PRIMITIVE
CONSTRUCTED = TypeNature.CONSTRUCTED


@pytest.mark.skip("TODO")
class TestBasics:
    def test_identifier_long(self):
        self.skipTest(
            "This is not yet implemented. I have not understood the "
            "spec to confidently write a test"
        )  # TODO


@pytest.mark.parametrize(
    "octet, expected_class, expected_pc, expected_value",
    [
        (0b00000010, UNIVERSAL, PRIMITIVE, 0b00010),
        (0b00100010, UNIVERSAL, CONSTRUCTED, 0b00010),
        (0b01000010, APPLICATION, PRIMITIVE, 0b00010),
        (0b01100010, APPLICATION, CONSTRUCTED, 0b00010),
        (0b10000010, CONTEXT, PRIMITIVE, 0b00010),
        (0b10100010, CONTEXT, CONSTRUCTED, 0b00010),
        (0b11000010, PRIVATE, PRIMITIVE, 0b00010),
        (0b11100010, PRIVATE, CONSTRUCTED, 0b00010),
    ],
)
def test_identifiers(octet, expected_class, expected_pc, expected_value):
    result = TypeInfo.from_bytes(octet)
    expected = TypeInfo(expected_class, expected_pc, expected_value)
    assert result == expected


@pytest.mark.parametrize(
    "typeclass, tag, pc, expected_class",
    [
        (UNIVERSAL, 0x00, PRIMITIVE, t.EOC),
        (UNIVERSAL, 0x01, PRIMITIVE, t.Boolean),
        (UNIVERSAL, 0x02, PRIMITIVE, t.Integer),
        (UNIVERSAL, 0x03, PRIMITIVE, t.BitString),
        (UNIVERSAL, 0x04, PRIMITIVE, t.OctetString),
        (UNIVERSAL, 0x05, PRIMITIVE, t.Null),
        (UNIVERSAL, 0x06, PRIMITIVE, t.ObjectIdentifier),
        (UNIVERSAL, 0x07, PRIMITIVE, t.ObjectDescriptor),
        (UNIVERSAL, 0x09, PRIMITIVE, t.Real),
        (UNIVERSAL, 0x0A, PRIMITIVE, t.Enumerated),
        (UNIVERSAL, 0x0C, PRIMITIVE, t.Utf8String),
        (UNIVERSAL, 0x0D, PRIMITIVE, t.RelativeOid),
        (UNIVERSAL, 0x12, PRIMITIVE, t.NumericString),
        (UNIVERSAL, 0x13, PRIMITIVE, t.PrintableString),
        (UNIVERSAL, 0x14, PRIMITIVE, t.T61String),
        (UNIVERSAL, 0x15, PRIMITIVE, t.VideotexString),
        (UNIVERSAL, 0x16, PRIMITIVE, t.IA5String),
        (UNIVERSAL, 0x17, PRIMITIVE, t.UtcTime),
        (UNIVERSAL, 0x18, PRIMITIVE, t.GeneralizedTime),
        (UNIVERSAL, 0x19, PRIMITIVE, t.GraphicString),
        (UNIVERSAL, 0x1A, PRIMITIVE, t.VisibleString),
        (UNIVERSAL, 0x1B, PRIMITIVE, t.GeneralString),
        (UNIVERSAL, 0x1C, PRIMITIVE, t.UniversalString),
        (UNIVERSAL, 0x1E, PRIMITIVE, t.BmpString),
        (UNIVERSAL, 0x08, CONSTRUCTED, t.External),
        (UNIVERSAL, 0x0B, CONSTRUCTED, t.EmbeddedPdv),
        (UNIVERSAL, 0x10, CONSTRUCTED, t.Sequence),
        (UNIVERSAL, 0x11, CONSTRUCTED, t.Set),
        (UNIVERSAL, 0x1D, CONSTRUCTED, t.CharacterString),
    ],
)
def test_class_detection(typeclass, tag, pc, expected_class):
    result = t.X690Type.get(typeclass, tag, pc)
    assert result == expected_class



# pylint: skip-file

import sys
from unittest import TestCase

import pytest

from x690.exc import IncompleteDecoding, UnexpectedType, X690Error
from x690.types import (
    Boolean,
    Integer,
    Null,
    ObjectIdentifier,
    OctetString,
    Sequence,
    T61String,
    UnknownType,
    X690Type,
    decode,
)
from x690.util import TypeInfo

from .conftest import assert_bytes_equal


class TestBoolean(TestCase):
    def test_encoding_false(self):
        value = Boolean(False)
        result = bytes(value)
        expected = b"\x01\x01\x00"
        assert_bytes_equal(result, expected)

    def test_encoding_true(self):
        value = Boolean(True)
        result = bytes(value)
        expected = b"\x01\x01\x01"
        assert_bytes_equal(result, expected)

    def test_decoding_false(self):
        result, _ = decode(b"\x01\x01\x00")
        expected = Boolean(False)
        self.assertEqual(result, expected)

    def test_decoding_true(self):
        result, _ = decode(b"\x01\x01\x01")
        expected = Boolean(True)
        self.assertEqual(result, expected)

        result, _ = decode(b"\x01\x01\x02")
        expected = Boolean(True)
        self.assertEqual(result, expected)

        result, _ = decode(b"\x01\x01\xff")
        expected = Boolean(True)
        self.assertEqual(result, expected)

    def test_pythonize(self):
        result = Boolean(True).pythonize()
        expected = True
        self.assertEqual(result, expected)

    def test_validate_too_long(self):
        """
        Validate what happens when there are too many bytes.
        """
        with self.assertRaisesRegex(ValueError, "Length"):
            Boolean.validate(b"\x01\x00\x00")


class TestObjectIdentifier(TestCase):
    def setUp(self):
        super().setUp()

    def test_uncollapsible_head(self):
        """
        OIDs need to be "collapsed" when converting to bytes. This requires the
        first two positions to be "small-enough"
        """
        with pytest.raises(ValueError) as exc:
            instance = ObjectIdentifier("200.200")
            bytes(instance)
        exc.match("Unable to collapse.*too large")

    def test_to_int(self):
        """
        An object-identifier of one node should be convertible to int
        """
        result = int(ObjectIdentifier("42"))
        assert result == 42

    def test_to_int_multiple(self):
        """
        An object-identifier of multiple node should not be convertible to int
        """
        with pytest.raises(ValueError) as exc:
            int(ObjectIdentifier("1.2.3"))
        exc.match("one node.*3 nodes")

    def test_raw_encoding_empty(self):
        """
        An empty object-identifier should be encodable
        """
        result = ObjectIdentifier().encode_raw()
        assert result == b""
        result = ObjectIdentifier(b"").encode_raw()
        assert result == b""

    def test_simple_encoding(self):
        """
        A simple OID with no identifier above 127
        """
        oid = ObjectIdentifier("1.3.6.1.2.1")
        result = bytes(oid)
        expected = b"\x06\x05\x2b\x06\x01\x02\x01"
        assert_bytes_equal(result, expected)

    def test_simple_decoding(self):
        """
        A simple OID with no identifier above 127
        """
        expected = ObjectIdentifier("1.3.6.1.2.1")
        result, _ = decode(b"\x06\x05\x2b\x06\x01\x02\x01")
        self.assertEqual(result, expected)

    def test_decoding_zero(self):
        """
        A simple OID with the top-level ID '0'
        """
        expected = ObjectIdentifier()
        result, _ = decode(b"\x06\x00")
        self.assertEqual(result, expected)

    def test_encoding_zero(self):
        """
        A simple OID with the top-level ID '0'
        """
        oid = ObjectIdentifier()
        result = bytes(oid)
        expected = b"\x06\x00"
        self.assertEqual(result, expected)

    def test_multibyte_encoding(self):
        """
        If a sub-identifier has a value bigger than 127, the encoding becomes a
        bit weird. The sub-identifiers are split into multiple sub-identifiers.
        """
        oid = ObjectIdentifier("1.3.6.8072")
        result = bytes(oid)
        expected = b"\x06\x04\x2b\x06\xbf\x08"
        assert_bytes_equal(result, expected)

    def test_multibyte_decoding(self):
        """
        If a sub-identifier has a value bigger than 127, the decoding becomes a
        bit weird. The sub-identifiers are split into multiple sub-identifiers.
        """
        expected = ObjectIdentifier("1.3.6.8072")
        result, _ = decode(b"\x06\x04\x2b\x06\xbf\x08")
        self.assertEqual(result, expected)

    def test_encode_large_value(self):
        """
        OID sub-identifiers larger than 127 must be split up.

        See https://en.wikipedia.org/wiki/Variable-length_quantity
        """
        result = ObjectIdentifier.encode_large_value(106903)
        expected = [0b10000110, 0b11000011, 0b00010111]
        self.assertEqual(result, expected)

    def test_fromstring(self):
        result = ObjectIdentifier("1.2.3")
        expected = ObjectIdentifier("1.2.3")
        self.assertEqual(result, expected)

    def test_fromstring_leading_dot(self):
        """
        A leading dot represents the "root" node. This should be allowed as
        string input.
        """
        result = ObjectIdentifier(".1.2.3")
        expected = ObjectIdentifier("1.2.3")
        self.assertEqual(result, expected)

    def test_pythonize(self):
        result = ObjectIdentifier("1.2.3").pythonize()
        expected = "1.2.3"
        self.assertEqual(result, expected)

    def test_str(self):
        result = str(ObjectIdentifier("1.2.3"))
        expected = "1.2.3"
        self.assertEqual(result, expected)

    def test_encode_root(self):
        result = bytes(ObjectIdentifier("1"))
        expected = b"\x06\x01\x01"
        assert_bytes_equal(result, expected)

    def test_containment_a(self):
        a = ObjectIdentifier("1.2.3.4")
        b = ObjectIdentifier("1.2.3")
        self.assertTrue(a in b)

    def test_containment_b(self):
        a = ObjectIdentifier("1.2.3.4")
        b = ObjectIdentifier("1.2.3.4")
        self.assertTrue(a in b)

    def test_containment_c(self):
        a = ObjectIdentifier("1.3.6.1.2.1.1.1.0")
        b = ObjectIdentifier("1.3.6.1.2.1")
        self.assertTrue(a in b)

    def test_non_containment_a(self):
        a = ObjectIdentifier("1.2.3")
        b = ObjectIdentifier("1.2.3.4")
        self.assertFalse(a in b)

    def test_non_containment_b(self):
        a = ObjectIdentifier("1.2.3.5")
        b = ObjectIdentifier("1.2.3.4")
        self.assertFalse(a in b)

    def test_non_containment_c(self):
        a = ObjectIdentifier("1.2.3.4")
        b = ObjectIdentifier("1.2.3.5")
        self.assertFalse(a in b)

    def test_non_containment_d(self):
        a = ObjectIdentifier("1.3.6.1.2.1.25.1.1.0")
        b = ObjectIdentifier("1.3.6.1.2.1.1.9")
        self.assertFalse(a in b)

    def test_non_containment_e(self):
        a = ObjectIdentifier("1.3.6.1.2.13")
        b = ObjectIdentifier("1.3.6.1.2.1")
        self.assertFalse(a in b)

    def test_create_by_iterable(self):
        result = ObjectIdentifier("1.2.3")
        expected = ObjectIdentifier("1.2.3")
        self.assertEqual(result, expected)

    def test_repr(self):
        result = repr(ObjectIdentifier("1.2.3"))
        expected = "ObjectIdentifier('1.2.3')"
        self.assertEqual(result, expected)

    def test_hash(self):
        """
        Test hash function and that it makes sense.
        """
        result = hash(ObjectIdentifier("1.2.3"))
        expected = hash(ObjectIdentifier("1.2.3"))
        self.assertEqual(result, expected)

    def test_non_containment_f(self):
        """
        This case showed up during development of bulk operations. Throwing it
        into the unit tests to ensure proper containment checks.
        """
        a = ObjectIdentifier("1.3.6.1.2.1.2.2.1.22")
        b = ObjectIdentifier("1.3.6.1.2.1.2.2.1.10.38")
        self.assertNotIn(a, b, "%s should not be in %s" % (a, b))
        self.assertNotIn(b, a, "%s should not be in %s" % (b, a))

    def test_length_1(self):
        """
        OIDs with one node should have a length of 1
        """
        obj = ObjectIdentifier("1")
        self.assertEqual(len(obj), 1)

    def test_length_ge1(self):
        """
        OIDs with more than one node should have a length equal to the number
        of nodes.
        """
        obj = ObjectIdentifier("1.2.3")
        self.assertEqual(len(obj), 3)

    def test_inequalitites(self):
        a = ObjectIdentifier("1.2.3")
        b = ObjectIdentifier("1.2.4")
        self.assertTrue(a < b)
        self.assertFalse(b < a)
        self.assertFalse(a < a)
        self.assertFalse(a > b)
        self.assertTrue(b > a)
        self.assertFalse(b > b)

    def test_concatenation(self):
        a = ObjectIdentifier("1.2.3")
        b = ObjectIdentifier("4.5.6")
        expected = ObjectIdentifier("1.2.3.4.5.6")
        result = a + b
        self.assertEqual(result, expected)

    def test_item_access(self):
        a = ObjectIdentifier("1.2.3")
        expected = 2
        result = a[1]
        self.assertEqual(result, expected)

    def test_slicing(self):
        a = ObjectIdentifier("1.2.3.4.5")
        expected = ObjectIdentifier("1.2.3.4")
        result = a[:-1]
        self.assertEqual(result, expected)

    def test_large_first_values(self):
        """
        When working with sub-trees the first two values can be larger than is
        accepted for *absolute* OIDs
        """
        str_value = "56.1342177480.31"
        oid = ObjectIdentifier(str_value)
        assert str_value == str(oid)


class TestInteger(TestCase):
    def test_encode_empty(self):
        result = Integer().encode_raw()
        assert result == b""

    def test_encoding(self):
        value = Integer(100)
        result = bytes(value)
        expected = b"\x02\x01\x64"
        assert_bytes_equal(result, expected)

    def test_decoding(self):
        result, _ = decode(b"\x02\x01\x0a")
        expected = Integer(10)
        self.assertEqual(result, expected)

    def test_encoding_large_value(self):
        value = Integer(1913359423)
        result = bytes(value)
        expected = b"\x02\x04\x72\x0b\x8c\x3f"
        assert_bytes_equal(result, expected)

    def test_decoding_large_value(self):
        result, _ = decode(b"\x02\x04\x72\x0b\x8c\x3f")
        expected = Integer(1913359423)
        self.assertEqual(result, expected)

    def test_encoding_zero(self):
        value = Integer(0)
        result = bytes(value)
        expected = b"\x02\x01\x00"
        assert_bytes_equal(result, expected)

    def test_decoding_zero(self):
        result, _ = decode(b"\x02\x01\x00")
        expected = Integer(0)
        self.assertEqual(result, expected)

    def test_decoding_minus_one(self):
        result, _ = decode(b"\x02\x01\xff")
        expected = Integer(-1)
        self.assertEqual(result, expected)

    def test_decoding_minus_large_value(self):
        result, _ = decode(b"\x02\x04\x8d\xf4\x73\xc1")
        expected = Integer(-1913359423)
        self.assertEqual(result, expected)

    def test_pythonize(self):
        result = Integer(1).pythonize()
        expected = 1
        self.assertEqual(result, expected)


class TestIntegerValues(TestCase):
    def test_32768(self):
        """
        Issue identified in github issue #27

        See https://github.com/exhuma/puresnmp/issues/27
        """
        value = Integer(32768)
        result = bytes(value)
        expected = b"\x02\x03\x00\x80\x00"
        assert_bytes_equal(result, expected)

    def test_minus_one(self):
        value = Integer(-1)
        result = bytes(value)
        expected = b"\x02\x01\xff"
        assert_bytes_equal(result, expected)

    def test_minus_two(self):
        value = Integer(-2)
        result = bytes(value)
        expected = b"\x02\x01\xfe"
        assert_bytes_equal(result, expected)

    def test_zero(self):
        value = Integer(0)
        result = bytes(value)
        expected = b"\x02\x01\x00"
        assert_bytes_equal(result, expected)

    def test_minus_16bit(self):
        value = Integer(-0b1111111111111111)
        result = bytes(value)
        expected = b"\x02\x03\xff\x00\x01"
        assert_bytes_equal(result, expected)

    def test_minus_16bit_plus_one(self):
        value = Integer(-0b1111111111111111 + 1)
        result = bytes(value)
        expected = b"\x02\x03\xff\x00\x02"
        assert_bytes_equal(result, expected)

    def test_minus_16bit_minus_one(self):
        value = Integer(-0b1111111111111111 - 1)
        result = bytes(value)
        expected = b"\x02\x03\xff\x00\x00"
        assert_bytes_equal(result, expected)

    def test_minus_16bit_minus_two(self):
        value = Integer(-0b1111111111111111 - 2)
        result = bytes(value)
        expected = b"\x02\x03\xfe\xff\xff"
        assert_bytes_equal(result, expected)

    def test_16bit(self):
        value = Integer(0b1111111111111111)
        result = bytes(value)
        expected = b"\x02\x03\x00\xff\xff"
        assert_bytes_equal(result, expected)

    def test_16bitplusone(self):
        value = Integer(0b1111111111111111 + 1)
        result = bytes(value)
        expected = b"\x02\x03\x01\x00\x00"
        assert_bytes_equal(result, expected)

    def test_16bitminusone(self):
        value = Integer(0b1111111111111111 - 1)
        result = bytes(value)
        expected = b"\x02\x03\x00\xff\xfe"
        assert_bytes_equal(result, expected)

    def test_32bit(self):
        value = Integer(0b11111111111111111111111111111111)
        result = bytes(value)
        expected = b"\x02\x05\x00\xff\xff\xff\xff"
        assert_bytes_equal(result, expected)


class TestOctetString(TestCase):
    def test_encoding(self):
        value = OctetString("hello")
        result = bytes(value)
        expected = b"\x04\x05hello"
        assert_bytes_equal(result, expected)

    def test_decoding(self):
        result, _ = decode(b"\x04\x05hello")
        expected = OctetString("hello")
        self.assertEqual(result, expected)

    def test_pythonize(self):
        result = OctetString("hello").pythonize()
        expected = b"hello"
        self.assertEqual(result, expected)

    def test_decoding_indef_length(self):
        data_def = b"\x04\x0bhello-world"
        data_indef = b"\x04\x80hello-world\x00\x00"
        result_def, _ = decode(data_def)
        result_indef, _ = decode(data_indef)
        assert result_def == result_indef


class TestT61String(TestCase):
    def test_encode_empty(self):
        result = T61String().encode_raw()
        assert result == b""

    def test_encoding(self):
        value = T61String("hello Ω")
        result = bytes(value)
        expected = b"\x14\x07hello \xe0"
        assert_bytes_equal(result, expected)

    def test_decoding(self):
        result, _ = decode(b"\x14\x07hello \xe0")
        expected = T61String("hello Ω")
        self.assertEqual(result, expected)

    def test_pythonize_from_string(self):
        obj = T61String("hello Ω")
        result = obj.pythonize()
        expected = "hello Ω"
        self.assertEqual(result, expected)

    def test_pythonize_from_bytes(self):
        obj = T61String(b"hello \xe0")
        result = obj.pythonize()
        expected = "hello Ω"
        self.assertEqual(result, expected)


class TestSequence(TestCase):
    def test_encode_raw(self):
        data = b"\x04\x05hello\x06\x02+\x06\x02\x01d"
        result = Sequence.decode_raw(data)
        expected = [
            OctetString(b"hello"),
            ObjectIdentifier("1.3.6"),
            Integer(100),
        ]
        assert result == expected

    def test_encode_raw_empty(self):
        result = Sequence().encode_raw()
        assert result == b""

    def test_eq_instancecheck(self):
        assert Sequence() != object()

    def test_encoding(self):
        value = Sequence(
            [OctetString("hello"), ObjectIdentifier("1.3.6"), Integer(100)]
        )
        result = bytes(value)
        expected = (
            bytes(
                [
                    0x30,
                    14,  # Expected length (note that an OID drops one byte)
                ]
            )
            + bytes(OctetString("hello"))
            + bytes(ObjectIdentifier("1.3.6"))
            + bytes(Integer(100))
        )
        assert_bytes_equal(result, expected)

    def test_raw_bytes_init(self):
        """
        After a normal init, the raw_bytes should be set
        """
        obj = Sequence(
            [
                Integer(1),
                Integer(2),
                OctetString("foo"),
            ]
        )
        self.assertEqual(obj.raw_bytes, b"\x02\x01\x01\x02\x01\x02\x04\x03foo")

    def test_decoding_simple(self):
        result, _ = decode(b"\x30\x0b\x02\x01\x01\x02\x01\x02\x04\x03foo")
        expected = Sequence(
            [
                Integer(1),
                Integer(2),
                OctetString("foo"),
            ]
        )
        self.assertEqual(result, expected)

    def test_decoding_recursive(self):
        result, _ = decode(
            b"\x30\x13"
            b"\x02\x01\x01"
            b"\x02\x01\x02"
            b"\x04\x03foo"
            b"\x30\x06"
            b"\x02\x01\x01"
            b"\x02\x01\x02"
        )
        expected = Sequence(
            [
                Integer(1),
                Integer(2),
                OctetString("foo"),
                Sequence(
                    [
                        Integer(1),
                        Integer(2),
                    ]
                ),
            ]
        )
        self.assertEqual(result, expected)

    def test_pythonize(self):
        result = Sequence(
            [Integer(1), Sequence([OctetString("123")])]
        ).pythonize()
        expected = [1, [b"123"]]
        self.assertEqual(result, expected)

    def test_iteration(self):
        data = Sequence(
            [Integer(1), Sequence([OctetString("123")]), OctetString(b"foo")]
        )
        result = [item for item in data]
        expected = [
            Integer(1),
            Sequence([OctetString("123")]),
            OctetString(b"foo"),
        ]
        self.assertEqual(result, expected)

    def test_indexing(self):
        data = Sequence([Integer(1), OctetString(b"foo")])
        result = data[1]
        expected = OctetString(b"foo")
        self.assertEqual(result, expected)

    def test_repr(self):
        result = repr(Sequence([Integer(10)]))
        expected = "Sequence([Integer(10)])"
        self.assertEqual(result, expected)

    def test_length_empty(self):
        result = len(Sequence())
        expected = 0
        self.assertEqual(result, expected)

    def test_length_nonempty(self):
        result = len(Sequence([Integer(1), Integer(2)]))
        expected = 2
        self.assertEqual(result, expected)


class TestNull(TestCase):
    def test_null_is_false(self):
        """
        The Null type should be considered as falsy.
        """
        self.assertFalse(Null())

    def test_validate_true(self):
        Null.validate(b"\x05\x00")

    def test_validate_false(self):
        with self.assertRaises(ValueError):
            Null.validate(b"\x05\x01")

    def test_encoding(self):
        result = bytes(Null())
        expected = b"\x05\x00"
        self.assertEqual(result, expected)

    def test_decode_null(self):
        expected = Null()
        result = Null.decode(b"\x05\x00\x00")
        self.assertEqual(result, expected)

    def test_repr(self):
        expected = "Null()"
        result = repr(Null())
        self.assertEqual(result, expected)


class TestUnknownType(TestCase):
    def test_decoding(self):
        result, _ = decode(b"\x99\x01\x0a")
        expected = UnknownType(b"\x0a", 0x99)
        self.assertEqual(result, expected)

    def test_encoding(self):
        result = bytes(UnknownType(b"\x0a", 0x99))
        expected = b"\xb9\x01\x0a"
        self.assertEqual(result, expected)

    def test_repr(self):
        result = repr(UnknownType(b"abc", 99))
        typeinfo = TypeInfo("application", "constructed", 3)
        expected = "<UnknownType 99 b'abc' application/constructed/3>"
        self.assertEqual(result, expected)

    def test_decoding_indef_length(self):
        data = bytearray(bytes(UnknownType(b"some-data", 0x99)))
        data_def = bytes(data)
        data[1] = 0x80
        data.extend([0x00, 0x00])
        data_indef = bytes(data)
        result_def, _ = decode(data_def)
        result_indef, _ = decode(data_indef)
        assert result_def == result_indef


class TestAllTypes(TestCase):
    """
    Tests which are valid for all types
    """

    def test_tlv_empty(self):
        with pytest.raises(IndexError):
            decode(b"")

    def test_tlv_simple(self):
        result = decode(bytes([2, 1, 0]))
        expected = (Integer(0), 3)
        self.assertEqual(result, expected)

    def test_tlv_unknown_type(self):
        result, next_index = decode(bytes([254, 1, 0]))
        expected = UnknownType(b"\x00", 254)
        self.assertEqual(next_index, 3)
        self.assertEqual(result, expected)
        self.assertEqual(result.tag, 254)
        self.assertEqual(result.length, 1)
        self.assertEqual(result.value, b"\x00")

    def test_validation_wrong_typeclass(self):
        with self.assertRaises(ValueError):
            Integer.validate(bytes([0b00111110]))

    def test_childof(self):
        a = ObjectIdentifier("1.2.3")
        b = ObjectIdentifier("1.2.3.1")
        c = ObjectIdentifier("1.2.4")
        d = ObjectIdentifier("1")
        self.assertTrue(b.childof(a))
        self.assertFalse(a.childof(b))
        self.assertTrue(a.childof(a))
        self.assertFalse(c.childof(a))
        self.assertFalse(a.childof(c))
        self.assertFalse(d.childof(c))
        self.assertTrue(c.childof(d))

    def test_parentdf(self):
        a = ObjectIdentifier("1.2.3")
        b = ObjectIdentifier("1.2.3.1")
        c = ObjectIdentifier("1.2.4")
        d = ObjectIdentifier("1")
        self.assertFalse(b.parentof(a))
        self.assertTrue(a.parentof(b))
        self.assertTrue(a.parentof(a))
        self.assertFalse(c.parentof(a))
        self.assertFalse(a.parentof(c))
        self.assertTrue(d.parentof(c))
        self.assertFalse(c.parentof(d))


@pytest.mark.parametrize("cls", X690Type.all())
def test_noarg_constructor(cls):
    """
    We want to allow a no-args constructor on everything. It should create
    an instance of the requested class
    """
    instance = cls()
    assert isinstance(instance, cls)


def test_enforcing_types():
    data = bytes(OctetString(b"foo"))
    with pytest.raises(UnexpectedType):
        decode(data, enforce_type=Integer)


def test_incomplete_decoding():
    """
    When calling decode in strict mode we want to raise an exception on
    unconsumed bytes.
    """
    data = bytes(OctetString(b"Hello")) + b"junk-bytes"
    with pytest.raises(IncompleteDecoding) as exc:
        decode(data, strict=True)
    assert exc.value.remainder == b"junk-bytes"


@pytest.mark.parametrize("cls", X690Type.all())
def test_repr(cls):
    obj = cls()
    result = repr(obj)
    assert cls.__name__ in result


def test_helpful_noarg_error():
    """
    Custom types with no "no-arg" constructor lead to cryptic errors if not
    dealt with. Ensure the error-message is somewhat helpful.
    """

    class Foo(X690Type[str]):
        def __init__(self, value: str) -> None:
            """
            This enforces *value* which breaks lazy-decoding
            """
            super().__init__(value)

    with pytest.raises(X690Error) as exc:
        Foo.from_bytes(b"hello")
    exc.match("no-arg.*x690.types.UNINITIALISED")



# pylint: skip-file

from unittest import TestCase

import pytest

from x690.util import (
    Length,
    TypeClass,
    TypeInfo,
    TypeNature,
    decode_length,
    encode_length,
    get_value_slice,
    visible_octets,
    wrap,
)

from .conftest import assert_bytes_equal


class TestTypeInfoDecoding(TestCase):
    """
    Tests the various possible combinations for decoding type-hint octets into
    Python objects.
    """

    def test_from_bytes_raw_value(self):
        result = TypeInfo.from_bytes(0b00011110)._raw_value
        expected = 0b00011110
        self.assertEqual(result, expected)

    def test_from_bytes_a(self):
        result = TypeInfo.from_bytes(0b00011110)
        expected = TypeInfo(TypeClass.UNIVERSAL, TypeNature.PRIMITIVE, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_b(self):
        result = TypeInfo.from_bytes(0b00111110)
        expected = TypeInfo(
            TypeClass.UNIVERSAL, TypeNature.CONSTRUCTED, 0b11110
        )
        self.assertEqual(result, expected)

    def test_from_bytes_c(self):
        result = TypeInfo.from_bytes(0b01011110)
        expected = TypeInfo(
            TypeClass.APPLICATION, TypeNature.PRIMITIVE, 0b11110
        )
        self.assertEqual(result, expected)

    def test_from_bytes_d(self):
        result = TypeInfo.from_bytes(0b01111110)
        expected = TypeInfo(
            TypeClass.APPLICATION, TypeNature.CONSTRUCTED, 0b11110
        )
        self.assertEqual(result, expected)

    def test_from_bytes_e(self):
        result = TypeInfo.from_bytes(0b10011110)
        expected = TypeInfo(TypeClass.CONTEXT, TypeNature.PRIMITIVE, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_f(self):
        result = TypeInfo.from_bytes(0b10111110)
        expected = TypeInfo(TypeClass.CONTEXT, TypeNature.CONSTRUCTED, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_g(self):
        result = TypeInfo.from_bytes(0b11011110)
        expected = TypeInfo(TypeClass.PRIVATE, TypeNature.PRIMITIVE, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_h(self):
        result = TypeInfo.from_bytes(0b11111110)
        expected = TypeInfo(TypeClass.PRIVATE, TypeNature.CONSTRUCTED, 0b11110)
        self.assertEqual(result, expected)


class TestTypeInfoEncoding(TestCase):
    """
    Tests the various possible combinations for encoding type-hint instances
    into bytes.
    """

    def test_to_bytes_a(self):
        obj = TypeInfo(TypeClass.UNIVERSAL, TypeNature.PRIMITIVE, 0b11110)
        result = bytes(obj)
        expected = bytes([0b00011110])
        self.assertEqual(result, expected)

    def test_to_bytes_b(self):
        obj = TypeInfo(TypeClass.UNIVERSAL, TypeNature.CONSTRUCTED, 0b11110)
        result = bytes(obj)
        expected = bytes([0b00111110])
        self.assertEqual(result, expected)

    def test_to_bytes_c(self):
        obj = TypeInfo(TypeClass.APPLICATION, TypeNature.PRIMITIVE, 0b11110)
        result = bytes(obj)
        expected = bytes([0b01011110])
        self.assertEqual(result, expected)

    def test_to_bytes_d(self):
        obj = TypeInfo(TypeClass.APPLICATION, TypeNature.CONSTRUCTED, 0b11110)
        result = bytes(obj)
        expected = bytes([0b01111110])
        self.assertEqual(result, expected)

    def test_to_bytes_e(self):
        obj = TypeInfo(TypeClass.CONTEXT, TypeNature.PRIMITIVE, 0b11110)
        result = bytes(obj)
        expected = bytes([0b10011110])
        self.assertEqual(result, expected)

    def test_to_bytes_f(self):
        obj = TypeInfo(TypeClass.CONTEXT, TypeNature.CONSTRUCTED, 0b11110)
        result = bytes(obj)
        expected = bytes([0b10111110])
        self.assertEqual(result, expected)

    def test_to_bytes_g(self):
        obj = TypeInfo(TypeClass.PRIVATE, TypeNature.PRIMITIVE, 0b11110)
        result = bytes(obj)
        expected = bytes([0b11011110])
        self.assertEqual(result, expected)

    def test_to_bytes_h(self):
        obj = TypeInfo(TypeClass.PRIVATE, TypeNature.CONSTRUCTED, 0b11110)
        result = bytes(obj)
        expected = bytes([0b11111110])
        self.assertEqual(result, expected)


class TestTypeInfoClass(TestCase):
    """
    Tests various "implied" functionality of TypeInfo objects.
    """

    def test_equality(self):
        a = TypeInfo(TypeClass.UNIVERSAL, TypeNature.CONSTRUCTED, 0b11110)
        b = TypeInfo(TypeClass.UNIVERSAL, TypeNature.CONSTRUCTED, 0b11110)
        self.assertEqual(a, b)

    def test_inequality(self):
        a = TypeInfo(TypeClass.UNIVERSAL, TypeNature.CONSTRUCTED, 0b11110)
        b = TypeInfo(TypeClass.UNIVERSAL, TypeNature.CONSTRUCTED, 0b10110)
        self.assertNotEqual(a, b)

    def test_encoding_symmetry_a(self):
        """
        Encoding an object to bytes, and then decoding the resulting bytes
        should yield the same instance.
        """
        expected = TypeInfo(
            TypeClass.UNIVERSAL, TypeNature.CONSTRUCTED, 0b11110
        )
        result = TypeInfo.from_bytes(bytes(expected))
        self.assertEqual(result, expected)

    def test_dencoding_symmetry_b(self):
        """
        Decoding an object from bytes, and then encoding the resulting instance
        should yield the same bytes.
        """
        expected = bytes([0b11111110])
        result = bytes(TypeInfo.from_bytes(expected))
        self.assertEqual(result, expected)

    def test_impossible_class(self):
        instance = TypeInfo(10, 100, 1000)
        with self.assertRaisesRegex(ValueError, "class"):
            bytes(instance)

    def test_impossible_pc(self):
        instance = TypeInfo(TypeClass.APPLICATION, 100, 1000)
        with self.assertRaisesRegex(ValueError, "primitive/constructed"):
            bytes(instance)


class TestLengthOctets(TestCase):
    def test_encode_length_short(self):
        expected = bytes([0b00100110])
        result = encode_length(38)
        self.assertEqual(result, expected)

    def test_encode_length_long(self):
        expected = bytes([0b10000001, 0b11001001])
        result = encode_length(201)
        assert_bytes_equal(result, expected)

    def test_encode_length_longer(self):
        expected = bytes([0b10000010, 0b00000001, 0b00101110])
        result = encode_length(302)
        assert_bytes_equal(result, expected)

    def test_encode_length_longer_2(self):
        expected = bytes([0x81, 0xA4])
        result = encode_length(164)
        assert_bytes_equal(result, expected)

    def test_encode_length_indefinite(self):
        expected = bytes([0b10000000])
        result = encode_length(Length.INDEFINITE)
        assert_bytes_equal(result, expected)

    def test_identifier_long(self):
        with self.assertRaises(NotImplementedError):
            TypeInfo.from_bytes(0b11111111)
        self.skipTest("Not yet implemented")  # TODO implement

    def test_decode_length_at_index(self):
        data = b"foobar\x05"
        expected = 5
        result, offset = decode_length(data, index=6)
        self.assertEqual(result, expected)
        self.assertEqual(offset, 1)

    def test_decode_length_short(self):
        data = b"\x05"
        expected = 5
        result, offset = decode_length(data)
        self.assertEqual(result, expected)
        self.assertEqual(offset, 1)

    def test_decode_length_long(self):
        data = bytes([0b10000010, 0b00000001, 0b10110011])
        expected = 435
        result, offset = decode_length(data)
        self.assertEqual(result, expected)
        self.assertEqual(offset, 3)

    def test_decode_length_longer(self):
        data = bytes([0x81, 0xA4])
        expected = 164
        result, offset = decode_length(data)
        self.assertEqual(result, expected)
        self.assertEqual(offset, 2)

    def test_decode_length_indefinite(self):
        data = bytes([0x80])
        expected = -1
        result, offset = decode_length(data)
        self.assertEqual(result, expected)
        self.assertEqual(offset, -1)

    def test_decode_length_reserved(self):
        with self.assertRaises(NotImplementedError):
            decode_length(bytes([0b11111111]))


class TestHelpers(TestCase):
    def test_visible_octets_minimal(self):
        result = visible_octets(bytes([0b00000000, 0b01010101]))
        expected = "00 55                                              .U"
        self.assertEqual(result, expected)

    def test_visible_octets_double_space(self):
        """
        Test that we have a double space after 8 octets for better readability
        """
        result = visible_octets(
            bytes(
                [
                    0b00000000,
                    0b01010101,
                    0b00000000,
                    0b01010101,
                    0b00000000,
                    0b01010101,
                    0b00000000,
                    0b01010101,
                    0b01010101,
                ]
            )
        )
        expected = (
            "00 55 00 55 00 55 00 55  55                        " ".U.U.U.UU"
        )
        self.assertEqual(result, expected)

    def test_visible_octets_multiline(self):
        """
        If we have more than 16 octets, we need to go to a new line.
        """
        result = visible_octets(bytes([0b00000000, 0b01010101] * 9))
        expected = (
            "00 55 00 55 00 55 00 55  00 55 00 55 00 55 00 55   "
            ".U.U.U.U.U.U.U.U\n"
            "00 55                                              "
            ".U"
        )
        self.assertEqual(result, expected)


class TestGithubIssue23(TestCase):
    """
    In issue #23 a problem was raised that the byte-order was incorrect when
    encoding large values for "length" information.

    This test-case takes the value (435) from the X.690 Wikipedia page as
    reference and has indeed highlighted the error.
    """

    def test_encode(self):
        expected = bytes([0b10000010, 0b00000001, 0b10110011])
        result = encode_length(435)
        assert_bytes_equal(result, expected)

    def test_decode(self):
        data = bytes([0b10000010, 0b00000001, 0b10110011])
        expected = 435
        result, offset = decode_length(data)
        self.assertEqual(result, expected)
        self.assertEqual(offset, 3)

    def test_symmetry(self):
        result, _ = decode_length(encode_length(435))
        self.assertEqual(result, 435)


def test_wrap():
    """
    Ensures that wrapping debug text works as expected
    """
    result = wrap("Hello\nThis is a long line", "Title", 3)
    expected = (
        "      ┌─────────────────────┐\n"
        "      │ Title               │\n"
        "      ├─────────────────────┤\n"
        "      │ Hello               │\n"
        "      │ This is a long line │\n"
        "      └─────────────────────┘"
    )
    assert result == expected


@pytest.mark.parametrize(
    "data, slc, next_index",
    [
        (b"\x02\01\01", slice(2, 3), 3),
        (b"\x04\03xxx", slice(2, 5), 5),
        (b"\x04\x80xxx\x00\x00", slice(2, 5), 7),
    ],
)
def test_value_slice(data, slc, next_index):
    res_slc, res_next_index = get_value_slice(data)
    assert res_slc == slc
    assert res_next_index == next_index


@pytest.mark.parametrize(
    "data, slc, next_index",
    [
        (b"padding\x02\x01\x01end-padding", slice(9, 10), 10),
        (b"padding\x04\x80the-value\x00\x00end-padding", slice(9, 18), 20),
    ],
)
def test_value_slice_indexed(data, slc, next_index):
    """
    We should be able to fetch a value slice starting at a given index
    """
    res_slc, res_next_index = get_value_slice(data, 7)
    assert res_slc == slc
    assert data[next_index:] == b"end-padding"



"""
Type instances should have raw-bytes easily available without the "type/length" header.
"""
import pytest

import x690.types as t


@pytest.mark.parametrize("cls", t.X690Type.all())
def test_raw_bytes(cls):
    try:
        instance = cls.decode(b"")
    except NotImplementedError:
        raise pytest.skip("Not yet implemented")
    assert instance.raw_bytes == b""


@pytest.mark.parametrize("cls", t.X690Type.all())
def test_raw_bytes(cls):
    instance = cls()
    assert isinstance(instance.raw_bytes, bytes)
